@@ -18,6 +18,7 @@
 #include <qpdf/QPDF.hh>
 #include <qpdf/QPDFExc.hh>
 #include <qpdf/Pl_QPDFTokenizer.hh>
+#include <qpdf/QIntC.hh>
 
 #include <stdexcept>
 
@@ -74,6 +75,12 @@ QPDF_Stream::unparse()
 	QUtil::int_to_string(this->generation) + " R";
 }
 
+JSON
+QPDF_Stream::getJSON()
+{
+    return this->stream_dict.getJSON();
+}
+
 QPDFObject::object_type_e
 QPDF_Stream::getTypeCode() const
 {
@@ -98,7 +105,8 @@ QPDF_Stream::setStreamDescription()
 {
     setDescription(
         this->qpdf,
-        "stream object " + QUtil::int_to_string(this->objid) + " " +
+        this->qpdf->getFilename() +
+        ", stream object " + QUtil::int_to_string(this->objid) + " " +
         QUtil::int_to_string(this->generation));
 }
 
@@ -127,11 +135,35 @@ QPDF_Stream::isDataModified() const
     return (! this->token_filters.empty());
 }
 
+qpdf_offset_t
+QPDF_Stream::getOffset() const
+{
+    return this->offset;
+}
+
+size_t
+QPDF_Stream::getLength() const
+{
+    return this->length;
+}
+
+PointerHolder<Buffer>
+QPDF_Stream::getStreamDataBuffer() const
+{
+    return this->stream_data;
+}
+
+PointerHolder<QPDFObjectHandle::StreamDataProvider>
+QPDF_Stream::getStreamDataProvider() const
+{
+    return this->stream_provider;
+}
+
 PointerHolder<Buffer>
 QPDF_Stream::getStreamData(qpdf_stream_decode_level_e decode_level)
 {
     Pl_Buffer buf("stream data buffer");
-    if (! pipeStreamData(&buf, 0, decode_level, false, false))
+    if (! pipeStreamData(&buf, nullptr, 0, decode_level, false, false))
     {
 	throw QPDFExc(qpdf_e_unsupported, qpdf->getFilename(),
                       "", this->offset,
@@ -145,7 +177,12 @@ PointerHolder<Buffer>
 QPDF_Stream::getRawStreamData()
 {
     Pl_Buffer buf("stream data buffer");
-    pipeStreamData(&buf, 0, qpdf_dl_none, false, false);
+    if (! pipeStreamData(&buf, nullptr, 0, qpdf_dl_none, false, false))
+    {
+	throw QPDFExc(qpdf_e_unsupported, qpdf->getFilename(),
+                      "", this->offset,
+                      "error getting raw stream data");
+    }
     QTC::TC("qpdf", "QPDF_Stream getRawStreamData");
     return buf.getBuffer();
 }
@@ -169,7 +206,7 @@ QPDF_Stream::understandDecodeParams(
             QPDFObjectHandle predictor_obj = decode_obj.getKey(key);
             if (predictor_obj.isInteger())
             {
-                predictor = predictor_obj.getIntValue();
+                predictor = predictor_obj.getIntValueAsInt();
                 if (! ((predictor == 1) || (predictor == 2) ||
                        ((predictor >= 10) && (predictor <= 15))))
                 {
@@ -186,7 +223,7 @@ QPDF_Stream::understandDecodeParams(
             QPDFObjectHandle earlychange_obj = decode_obj.getKey(key);
             if (earlychange_obj.isInteger())
             {
-                int earlychange = earlychange_obj.getIntValue();
+                int earlychange = earlychange_obj.getIntValueAsInt();
                 early_code_change = (earlychange == 1);
                 if (! ((earlychange == 0) || (earlychange == 1)))
                 {
@@ -205,7 +242,7 @@ QPDF_Stream::understandDecodeParams(
             QPDFObjectHandle param_obj = decode_obj.getKey(key);
             if (param_obj.isInteger())
             {
-                int val = param_obj.getIntValue();
+                int val = param_obj.getIntValueAsInt();
                 if (key == "/Columns")
                 {
                     columns = val;
@@ -361,6 +398,10 @@ QPDF_Stream::filterable(std::vector<std::string>& filters,
 
     QPDFObjectHandle decode_obj = this->stream_dict.getKey("/DecodeParms");
     std::vector<QPDFObjectHandle> decode_parms;
+    if (decode_obj.isArray() && (decode_obj.getArrayNItems() == 0))
+    {
+        decode_obj = QPDFObjectHandle::newNull();
+    }
     if (decode_obj.isArray())
     {
         for (int i = 0; i < decode_obj.getArrayNItems(); ++i)
@@ -431,8 +472,8 @@ QPDF_Stream::filterable(std::vector<std::string>& filters,
 }
 
 bool
-QPDF_Stream::pipeStreamData(Pipeline* pipeline,
-                            unsigned long encode_flags,
+QPDF_Stream::pipeStreamData(Pipeline* pipeline, bool* filterp,
+                            int encode_flags,
                             qpdf_stream_decode_level_e decode_level,
                             bool suppress_warnings, bool will_retry)
 {
@@ -444,7 +485,14 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
     bool early_code_change = true;
     bool specialized_compression = false;
     bool lossy_compression = false;
-    bool filter = (! ((encode_flags == 0) && (decode_level == qpdf_dl_none)));
+    bool ignored;
+    if (filterp == nullptr)
+    {
+        filterp = &ignored;
+    }
+    bool& filter = *filterp;
+    filter = (! ((encode_flags == 0) && (decode_level == qpdf_dl_none)));
+    bool success = true;
     if (filter)
     {
 	filter = filterable(filters, specialized_compression, lossy_compression,
@@ -469,6 +517,7 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
     if (pipeline == 0)
     {
 	QTC::TC("qpdf", "QPDF_Stream pipeStreamData with null pipeline");
+        // Return value is whether we can filter in this case.
 	return filter;
     }
 
@@ -516,7 +565,9 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
                     QTC::TC("qpdf", "QPDF_Stream PNG filter");
                     pipeline = new Pl_PNGFilter(
                         "png decode", pipeline, Pl_PNGFilter::a_decode,
-                        columns, colors, bits_per_component);
+                        QIntC::to_uint(columns),
+                        QIntC::to_uint(colors),
+                        QIntC::to_uint(bits_per_component));
                     to_delete.push_back(pipeline);
                 }
                 else if (predictor == 2)
@@ -524,7 +575,9 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
                     QTC::TC("qpdf", "QPDF_Stream TIFF predictor");
                     pipeline = new Pl_TIFFPredictor(
                         "tiff decode", pipeline, Pl_TIFFPredictor::a_decode,
-                        columns, colors, bits_per_component);
+                        QIntC::to_uint(columns),
+                        QIntC::to_uint(colors),
+                        QIntC::to_uint(bits_per_component));
                     to_delete.push_back(pipeline);
                 }
             }
@@ -585,8 +638,21 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
     else if (this->stream_provider.getPointer())
     {
 	Pl_Count count("stream provider count", pipeline);
-	this->stream_provider->provideStreamData(
-	    this->objid, this->generation, &count);
+        if (this->stream_provider->supportsRetry())
+        {
+            if (! this->stream_provider->provideStreamData(
+                    this->objid, this->generation, &count,
+                    suppress_warnings, will_retry))
+            {
+                filter = false;
+                success = false;
+            }
+        }
+        else
+        {
+            this->stream_provider->provideStreamData(
+                this->objid, this->generation, &count);
+        }
 	qpdf_offset_t actual_length = count.getCount();
 	qpdf_offset_t desired_length = 0;
         if (this->stream_dict.hasKey("/Length"))
@@ -634,6 +700,7 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
                                          will_retry))
         {
             filter = false;
+            success = false;
         }
     }
 
@@ -664,7 +731,7 @@ QPDF_Stream::pipeStreamData(Pipeline* pipeline,
                      " for content normalization in the manual."));
     }
 
-    return filter;
+    return success;
 }
 
 void
@@ -710,7 +777,8 @@ QPDF_Stream::replaceFilterData(QPDFObjectHandle const& filter,
     else
     {
         this->stream_dict.replaceKey(
-            "/Length", QPDFObjectHandle::newInteger(length));
+            "/Length", QPDFObjectHandle::newInteger(
+                QIntC::to_longlong(length)));
     }
 }
 
@@ -722,7 +790,7 @@ QPDF_Stream::replaceDict(QPDFObjectHandle new_dict)
     QPDFObjectHandle length_obj = new_dict.getKey("/Length");
     if (length_obj.isInteger())
     {
-        this->length = length_obj.getIntValue();
+        this->length = QIntC::to_size(length_obj.getUIntValue());
     }
     else
     {
